package com.pseuco

import kotlinx.serialization.Decoder
import kotlinx.serialization.Encoder
import kotlinx.serialization.KSerializer
import kotlinx.serialization.Required
import kotlinx.serialization.SerialDescriptor
import kotlinx.serialization.Serializable
import kotlinx.serialization.UnionKind
import kotlinx.serialization.builtins.serializer

/**
 * Class representing sharable pseuCo.com files.
 *
 * @author Konstantin Kopper
 * @since 2.0.0
 * @property name The name of the file.
 * @property content The actual body of the file.
 * @property type The file type.
 */
@Serializable
data class PseuCoComFile(val name: String, val content: String, @Required val type: Type = Type.PSEUCO) {

    /**
     * Enum representing the file types supported by the sharing API.
     */
    @Serializable(with = Type.Companion::class)
    enum class Type {

        /**
         * A pseuCo file.
         */
        PSEUCO,

        /**
         * A CCS file.
         */
        CCS,

        /**
         * A LTS file.
         */
        LTS;

        /**
         * Serializer converting enum values to lower case strings.
         *
         * @author Konstantin Kopper
         * @since 2.0.4
         */
        companion object : KSerializer<Type> {
            override val descriptor = SerialDescriptor("PseuCoComFile.Type", UnionKind.ENUM_KIND) {
                values().forEach { element(it.name.toLowerCase(), String.serializer().descriptor) }
            }

            override fun deserialize(decoder: Decoder): Type = values()[descriptor.getElementIndex(decoder.decodeString())]

            override fun serialize(encoder: Encoder, value: Type) = encoder.encodeString(descriptor.getElementName(value.ordinal))
        }
    }
}
