package com.pseuco.websocket

import com.pseuco.PseuCoComFile
import kotlinx.serialization.CompositeDecoder
import kotlinx.serialization.Decoder
import kotlinx.serialization.Encoder
import kotlinx.serialization.KSerializer
import kotlinx.serialization.SerialDescriptor
import kotlinx.serialization.Serializable
import kotlinx.serialization.SerializationException
import kotlinx.serialization.UnionKind
import kotlinx.serialization.builtins.serializer

/**
 * All valid messages used in the pseuCo WebSockets file sharing protocol.
 * See [this GitLab issue](https://dgit.cs.uni-saarland.de/pseuco/pseuco-ide/issues/15) for more information.
 *
 * @author Konstantin Kopper
 * @since 2.0.0
 * @see PseuCoWebSocket
 * @param type A string representation of the message type.
 * @constructor Creates a new Socket message.
 */
@Serializable(with = SocketMessage.Companion::class)
sealed class SocketMessage(val type: String) {

    /**
     * All valid request messages.
     *
     * @author Konstantin Kopper
     * @since 2.0.0
     * @constructor Creates a new request with [type].
     */
    sealed class Request(type: String) : SocketMessage(type) {

        /**
         * A request of [Request.type] `get`.
         * Holds a unique [id].
         * Should be answered with a [Response.File] message.
         *
         * @author Konstantin Kopper
         * @since 2.0.0
         * @param id Unique identifier of the requested file.
         */
        data class Get(val id: Int) : Request("get")

        /**
         * A request of [Request.type] `open`.
         * Holds a [file], which should be stored locally and opened in the IDE.
         * Should be answered with a [Response.Success] message.
         *
         * @author Konstantin Kopper
         * @since 2.0.0
         * @param file The file to be opened.
         */
        data class Open(val file: PseuCoComFile) : Request("open")
    }

    /**
     * All valid response messages.
     *
     * @author Konstantin Kopper
     * @since 2.0.0
     * @constructor Creates a new response with [type].
     */
    sealed class Response(type: String) : SocketMessage(type) {

        /**
         * A response of [Response.type] `file`.
         * Holds the requested [file].
         * Should follow on a [Request.Get] message.
         *
         * @author Konstantin Kopper
         * @since 2.0.0
         * @param file The requested file.
         */
        data class File(val file: PseuCoComFile) : Response("file")

        /**
         * A response of [Response.type] `success`.
         * Should follow on a [Request.Open] message.
         *
         * @author Konstantin Kopper
         * @since 2.0.0
         */
        object Success : Response("success")

        /**
         * A response of [Response.type] `error`.
         * Holds a [message] describing the error, and a [code] which may identify the error precisely.
         * May follow on any [Request] message.
         *
         * @author Konstantin Kopper
         * @since 2.0.0
         * @param message A description of the error.
         * @param code An identifier which may identify the error precisely.
         */
        data class Error(val message: String, val code: Code = Code.GENERAL) : Response("error") {

            /**
             * Possible error codes used in the [Error] messages.
             *
             * @author Konstantin Kopper
             */
            enum class Code {

                /**
                 * Default code describing arbitrary error messages.
                 *
                 * @author Konstantin Kopper
                 * @since 2.0.0
                 */
                GENERAL,

                /**
                 * Error following a [Request.Get], where no file was found for [Request.Get.id].
                 *
                 * @author Konstantin Kopper
                 * @since 2.0.0
                 */
                FILE_NOT_FOUND;

                /**
                 * Serializer for error [Code]s. Converts enum values to camel case strings.
                 *
                 * @author Konstantin Kopper
                 * @since 2.0.4
                 */
                companion object : KSerializer<Code> {
                    private val stringValues = values().map {
                        it.name.toLowerCase().replace(Regex("_(\\w)")) { match -> match.groupValues[1].toUpperCase() }
                    }

                    override val descriptor = SerialDescriptor("SocketMessage.Response.Error.Code", UnionKind.ENUM_KIND) {
                        values().forEach { element(it.name, String.serializer().descriptor) }
                    }

                    override fun deserialize(decoder: Decoder): Code = values()[stringValues.indexOf(decoder.decodeString())]

                    override fun serialize(encoder: Encoder, value: Code) = encoder.encodeString(stringValues[value.ordinal])
                }
            }
        }
    }

    /**
     * Serializer for [SocketMessage]s.
     *
     * @author Konstantin Kopper
     * @since 2.0.4
     */
    companion object : KSerializer<SocketMessage> {
        override val descriptor = SerialDescriptor("SocketMessage") {
            element("type", String.serializer().descriptor)
            element("id", Int.serializer().descriptor, isOptional = true)
            element("file", PseuCoComFile.serializer().descriptor, isOptional = true)
            element("message", String.serializer().descriptor, isOptional = true)
            element("code", Response.Error.Code.descriptor, isOptional = true)
        }

        override fun deserialize(decoder: Decoder): SocketMessage {
            val structure = decoder.beginStructure(descriptor)

            lateinit var type: String
            var id: Int? = null
            lateinit var file: PseuCoComFile
            lateinit var message: String
            lateinit var code: Response.Error.Code

            tailrec fun decodeAttributes(): SocketMessage {
                when (val i = structure.decodeElementIndex(descriptor)) {
                    CompositeDecoder.READ_DONE -> {
                        structure.endStructure(descriptor)
                        return when (type) {
                            "get" -> Request.Get(id!!)
                            "open" -> Request.Open(file)
                            "file" -> Response.File(file)
                            "success" -> Response.Success
                            "error" -> Response.Error(message, code)
                            else -> throw SerializationException("Unknown message type: '$type'.")
                        }
                    }
                    0 -> type = structure.decodeStringElement(descriptor, i)
                    1 -> id = structure.decodeIntElement(descriptor, i)
                    2 -> file = structure.decodeSerializableElement(descriptor, i, PseuCoComFile.serializer())
                    3 -> message = structure.decodeStringElement(descriptor, i)
                    4 -> code = structure.decodeSerializableElement(descriptor, i, Response.Error.Code.Companion)
                }

                return decodeAttributes()
            }

            return decodeAttributes()
        }

        override fun serialize(encoder: Encoder, value: SocketMessage) {
            val structure = encoder.beginStructure(descriptor)
            structure.encodeStringElement(descriptor, 0, value.type)
            when (value) {
                is Request.Get -> structure.encodeIntElement(descriptor, 1, value.id)
                is Request.Open -> structure.encodeSerializableElement(descriptor, 2, PseuCoComFile.serializer(), value.file)
                is Response.File -> structure.encodeSerializableElement(descriptor, 2, PseuCoComFile.serializer(), value.file)
                is Response.Success -> {
                }
                is Response.Error -> {
                    structure.encodeStringElement(descriptor, 3, value.message)
                    structure.encodeSerializableElement(descriptor, 4, Response.Error.Code.Companion, value.code)
                }
            }
            structure.endStructure(descriptor)
        }
    }
}
